#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : LGBSelector
# @Time         : 2020/10/13 7:34 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import pandas as pd
from tqdm.auto import tqdm

from tql.algo_ml.cv import LGBMClassifierCV


class FS(object):

    def __init__(self, sorted_feats_by_feature_importances, max_features=100, early_stopping_steps=-1, params=None,
                 cv=5, cv_seed=777):
        self.sorted_feats = sorted_feats_by_feature_importances
        self.max_features = max_features
        self.early_stopping_steps = early_stopping_steps
        self.clf = LGBMClassifierCV(params, cv, cv_seed)
        self.df_oofs = pd.DataFrame()
        self.selected_features_ = []

    def fit(self, train_df, y, test_df=None):
        step = 0
        score = 0
        self.scores = []
        print(f"Selecting until scores don't improve for {self.early_stopping_steps} rounds.")

        tqdm_ = tqdm(list(enumerate(self.sorted_feats)))
        for i, feat in tqdm_:
            tqdm_.set_description(feat)
            self.selected_features_.append(feat)

            self.clf.fit(
                train_df[self.selected_features_].values, y,
                test_df[self.selected_features_].values if test_df is not None else None,
                early_stopping_rounds=64, verbose=0
            )

            if self.clf.oof_score > score:
                step = 0  # 置0
                score = self.clf.oof_score
                self.scores.append(score)  # 历史得分方便画图

                if test_df is not None:
                    self.df_oofs[f"oof_{i}"] = self.clf.oof_train_and_test  # 方便作stacking

            else:
                self.selected_features_.pop(-1)
                step += 1

                if len(self.selected_features_) == self.max_features:
                    break

                if self.early_stopping_steps > 0:
                    if step == self.early_stopping_steps:  # 累计n个特征未提升终止特征选择
                        break



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : utils
# @Time         : 2021/9/15 上午10:11
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  :

import torch


def define_device(device_name):
    """
    Define the device to use during training and inference.
    If auto it will detect automatically whether to use cuda or cpu

    Parameters
    ----------
    device_name : str
        Either "auto", "cpu" or "cuda"

    Returns
    -------
    str
        Either "cpu" or "cuda"
    """
    if device_name == "auto":
        if torch.cuda.is_available():
            return "cuda"
        else:
            return "cpu"
    elif device_name == "cuda" and not torch.cuda.is_available():
        return "cpu"
    else:
        return device_name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : eda
# @Time         : 2021/9/30 下午6:46
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  : 探索性数据分析（Exploratory Data Analysis，EDA）


"""https://www.cnblogs.com/HuZihu/p/11146493.html
1. summary
2. train/test/目标值可视化分析:
    1）train与test分布是否一致【判别器、异常值、聚类】 oof分布
    2）train特征与label的关系
3. 缺失值分析: missingno 与 label的关系
4. 相关性分析
5. 异常值分析: 独立森林、箱型图、与label的关系
"""

import missingno
from pandas_summary import DataFrameSummary

# ME
from meutils.pipe import *


class EDA(object):

    def __init__(self, df):
        self.dfs = DataFrameSummary(df)

    def summary(self, transpose=False):
        _ = self.dfs.summary()
        return _ if transpose is False else _.T

    def missingno_bar(self, **kwargs):
        missingno.bar(self.dfs.df, **kwargs)

    def missingno_matrix(self, **kwargs):
        missingno.matrix(self.dfs.df, **kwargs)

    def missingno_heatmap(self, **kwargs):
        missingno.heatmap(self.dfs.df, **kwargs)

    def profiling(self, title="Pandas Profiling Report", **kwargs):
        from pandas_profiling import ProfileReport

        self.profile_report = ProfileReport(self.dfs.df, title=title, lazy=False, **kwargs)
        return self.profile_report

    # from functools import partialmethod
    # set_alive = partialmethod(missingno_bar, df)


if __name__ == '__main__':
    from sklearn.datasets import load_iris

    dfs = load_iris(return_X_y=True, as_frame=True)
    df = pd.concat(dfs, axis=1)

    eda = EDA(df)

    print(eda.summary())
    print(eda.summary(1))
    eda.missingno_heatmap()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : baidu
# @Time         : 2021/9/13 下午4:36
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  : https://ai.baidu.com/ai-doc/IMAGESEARCH/Pk3bczxau

"""
client.commentTag
client.depParser
client.dnnlm
client.ecnet
client.emotion
client.keyword
client.lexer
client.lexerCustom
client.newsSummary
client.sentimentClassify
client.simnet
client.topic
client.wordEmbedding
client.wordSimEmbedding
"""

from aip import AipNlp

keys = """
15696121
md9ONR0cj5pvF9oxYlg9MIMg
iYGkDYuW4XDGXjMtXPZclUplLhiBNBgQ
"""
client = AipNlp(*keys.split())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : base
# @Time         : 2021/9/16 上午10:12
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  :

import optuna
from optuna.samplers import TPESampler, CmaEsSampler

from meutils.pipe import *

optuna.logging.set_verbosity(optuna.logging.ERROR)


class Tuner(object):
    """https://github.com/optuna/optuna-examples"""

    def __init__(self, search_space, **kwargs):
        """https://github.com/optuna/optuna-examples

        @param search_space: Union[str, dict]
        @param kwargs:
        """
        self.search_space = self.search_space_from_yaml(search_space)
        self.kwargs = kwargs

    @abstractmethod
    def objective(self, trial: optuna.trial.Trial):
        """
        X = ...
        y = ...
        feval = roc_auc_score

        class LGBOptimizer(Tuner):
            def objective(self, trial: optuna.trial.Trial):
                params = self.trial_choice(trial)
                _ = LGBMClassifier(params).run(X, y, feval=feval)
                return _
        """

        params = self.trial_choice(trial)

        raise NotImplementedError("overwrite objective!!!")

    def optimize(self, trials=3,
                 direction='maximize',
                 study_name='optimizer',
                 sampler=TPESampler(seed=222),
                 pruner=None,
                 gc_after_trial=False,
                 storage="sqlite:///opt.db",
                 load_if_exists=True
                 ):

        self.study = optuna.create_study(
            study_name=study_name,
            direction=direction,
            sampler=sampler,
            pruner=pruner,
            storage=storage.replace(storage, f'sqlite:///opt_{study_name}.db'),
            load_if_exists=load_if_exists
        )

        self.study.optimize(
            self.objective,
            n_trials=trials,
            gc_after_trial=gc_after_trial,
            show_progress_bar=True
        )

        return self.study.best_params

    def plot_history(self):
        # optuna.visualization.plot_slice(self.study)
        _ = optuna.visualization.plot_optimization_history(self.study)

        # optuna.visualization.plot_param_importances(self.study)
        # optuna.visualization.plot_edf(self.study)
        # optuna.visualization.plot_parallel_coordinate(self.study)
        return _

    @property
    def trials_dataframe(self):
        df_trials = self.study.trials_dataframe().sort_values('value', ascending=False)

        return df_trials

    def top_params(self, topK=5, save_file=None):
        params_dict = (
            self.trials_dataframe.iloc[:topK, :]
                .filter(like='params_')
                .rename(columns=lambda col: col.split('params_')[1])
                .to_dict('records')
        )
        if save_file is not None:
            with open(save_file, 'w') as f:
                print(params_dict, file=f)

        return params_dict

    @staticmethod
    def search_space_from_yaml(search_space: Union[str, dict]):
        if isinstance(search_space, dict):
            return search_space
        elif isinstance(search_space, str) and Path(search_space).is_file():
            with open(search_space) as f:
                return yaml.safe_load(f)
        else:
            raise Exception('load params error')

    def trial_choice(self, trial: optuna.trial.Trial):
        params = {}
        for k, v in self.search_space.items():
            if isinstance(v, dict):
                v = v.copy()
                suggest_type = v.pop('type') if 'type' in v else v.pop('suggest_type')

                choice_func = trial.__getattribute__(f"suggest_{suggest_type}")
                params[k] = choice_func(k, **v)
            # elif isinstance(v, (int, float, bool, List[int], List[float])):
            elif isinstance(v, (str, int, float, bool, list)) or v is None:
                # choices = v if isinstance(v, (List[int], List[float], List[bool])) else [v]
                choices = v if isinstance(v, list) else [v]
                params[k] = trial.suggest_categorical(k, choices)  # [None]

            else:
                raise TypeError(f'suggest_type error: {k}_{v}')
        return params



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : Features
# @Time         : 2019-07-26 10:04
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : featuretools

import featuretools
import featuretools as ft
import featuretools.variable_types as vt

from tqdm.auto import tqdm
from tql.utils.pandas_utils import reduce_mem_usage, duplicate_columns
from tql.ml.automl.primitives import *
from featuretools.selection import remove_low_information_features, remove_single_value_features, \
    remove_highly_correlated_features


class AutoFeat(object):

    def __init__(self, id='ES'):
        self.es = ft.EntitySet(id=id)

    def add_entity(
            self, df, entity_id, type2features=None,
            index=None, time_index=None, secondary_time_index=None,
            normalize_entity=True,
    ):
        """

        :param df:
        :param entity_id: 表名
        :param type2features: vt.Categorical, vt.Boolean, vt.Id
        :param index:
        :param time_index:
        :param secondary_time_index:
        :return:
        """
        index = '__id' if index is None else index
        type2features = type2features if type2features is not None else {}

        variable_types = self._get_variable_types(self._convert_type(type2features))

        self.es.entity_from_dataframe(
            entity_id=entity_id,
            dataframe=df,
            index=index,
            variable_types=variable_types,
            time_index=time_index,
            secondary_time_index=secondary_time_index
        )

        if normalize_entity:
            self._normalize_entity(entity_id, variable_types)

    def add_relationship(self, parent_variable_name, child_variable_name):
        """
            relation = ft.Relationship(es['t1']['id'], es['t2']['id'])
            es.add_relationship(relation)
        :param parent_variable_name: (entity_id, id)
        :param child_variable_name: (entity_id, id)
        :return:
        """
        # relation = ft.Relationship(parent_variable, child_variable)
        relation = ft.Relationship(
            parent_variable=self.es[parent_variable_name[0]][parent_variable_name[1]],
            child_variable=self.es[child_variable_name[0]][child_variable_name[1]]
        )
        self.es.add_relationship(relation)

    def run_dfs(self,
                target_entity_id,
                max_depth=1,
                features_only=True,
                ignore_variables=None,
                reduce_mem=False,
                reduce_feats=True,
                trans_primitives=None,
                agg_primitives=None,
                chunk_size=None,
                n_jobs=1,
                **kwargs
                ):
        """Deep Feature Synthesisf
        agg_primitives (list[str or AggregationPrimitive], optional): List of Aggregation
            Feature types to apply.

                Default: ["sum", "std", "max", "skew", "min", "mean", "count", "percent_true", "num_unique", "mode"]
                DateTime: ['time_since_last', 'time_since_first', 'trend']

        trans_primitives (list[str or TransformPrimitive], optional):
            List of Transform Feature functions to apply.

                Default: ["day", "year", "month", "weekday", "haversine", "num_words", "num_characters"]

        groupby_trans_primitives (list[str or :class:`.primitives.TransformPrimitive`], optional):
            list of Transform primitives to make GroupByTransformFeatures with

        """
        if ignore_variables is None:
            # ignore_variables = [self.target_entity_id, self.index]
            # ignore_variables = ["__id"]  # 忽略单值id 会少了一些count特征
            ignore_variables = []

        if trans_primitives is None:
            trans_primitives = [
                "year", "month", "day", "hour", "minute", "week", "weekday", "is_weekend",
                'time_since_previous',
                # diff # https://stackoverflow.com/questions/60324672/how-is-time-since-previous-computed-in-featuretools
                Quarter(),
            ]

        _ = ft.dfs(
            entityset=self.es,
            target_entity=target_entity_id,  # 具有唯一ID: 不重复id的base_es或者normalize_entity生成的唯一id es
            features_only=features_only,
            max_depth=max_depth,
            # ignore_variables={self.entity_id: ignore_variables},
            chunk_size=chunk_size,
            n_jobs=n_jobs,
            verbose=1,
            agg_primitives=agg_primitives,
            trans_primitives=trans_primitives,
            **kwargs
        )

        if features_only:
            return _
        else:
            df = _[0]

            if reduce_feats:
                df = self._reduce_feats(df)

            if reduce_mem:
                df = reduce_mem_usage(df)

            df_bool = df.select_dtypes(bool)
            df.loc[:, df_bool.columns] = df_bool.astype(int)

            df_object = df.select_dtypes(object)
            df.loc[:, df_object.columns] = df_object.astype(float)

            return df.reset_index()

    def _reduce_feats(self, df):
        df = remove_low_information_features(df)
        df = remove_single_value_features(df, count_nan_as_value=True)
        df.drop(duplicate_columns(df), 1, inplace=True)
        return df

    def _normalize_entity(self, entity_id, variable_types):
        normalize_entity_feats = []
        for feat, type_ in variable_types.items():
            if type_ in {vt.Id, vt.Categorical, vt.Boolean} and feat != self.es[entity_id].index:
                normalize_entity_feats.append(feat)

        with tqdm(normalize_entity_feats) as t:
            for feat in t:
                t.set_description(f"NormalizeEntity[{feat}]")
                self.es.normalize_entity(entity_id, feat, feat)

    def _get_variable_types(self, type2features):
        dic = {}
        for type_, features in type2features.items():
            dic.update(zip(features, len(features) * [type_]))
        return dic

    def _convert_type(self, type2features):
        type2features_ = {}
        for type_, features in type2features.items():
            if isinstance(type_, str):
                type2features_[vt.__getattribute__(type_)] = features
            else:
                type2features_[type_] = features

        return type2features_


if __name__ == '__main__':
    import pandas as pd

    df = pd.DataFrame([[1, 2, 3], [2, 2, 3], [3, 2, 3]], columns=['uid', 'a', 'b'])
    df['date'] = pd.date_range('2020-01-01', periods=3)

    type2features = {
        'Id': ['uid'],
        'Categorical': ['a', 'b'],
        'Datetime': ['date']
    }

    af = AutoFeat()
    af.add_entity(df, 'demo', type2features, index='uid')
    # af.add_relationship((), ())

    df_ft = af.run_dfs('demo', 3, False, reduce_feats=False)

    print(df_ft)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : feature_selector
# @Time         : 2021/9/30 下午6:43
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  :

"""https://www.cnblogs.com/nolonely/p/6435083.html
1、为什么要做特征选择
在有限的样本数目下，用大量的特征来设计分类器计算开销太大而且分类性能差。
2、特征选择的确切含义
将高维空间的样本通过映射或者是变换的方式转换到低维空间，达到降维的目的，然后通过特征选取删选掉冗余和不相关的特征来进一步降维。
3、特征选取的原则
获取尽可能小的特征子集，不显著降低分类精度、不影响类分布以及特征子集应具有稳定适应性强等特点
"""

import scipy as sp

from functools import partial
from sklearn.utils import shuffle, check_random_state
from sklearn.preprocessing import scale as zscore
from sklearn.model_selection import train_test_split

# ME
from meutils.pipe import *
from aizoo.utils.model_utils import get_imp
from aizoo.utils.check_utils import check_classification


class SimpleFS(object):
    """粗筛
    高缺失率：
    低方差（高度重复值）：0.5%~99.5%分位数内方差为0的初筛
    高相关：特别高的初筛，根据重要性细筛 TODO
    低重要性：
    召回高IV：

    """

    def __init__(self, df: pd.DataFrame, exclude=None):
        self.to_drop_dict = {}
        if exclude is None:
            exclude = []

        assert isinstance(exclude, list)
        assert isinstance(df, pd.DataFrame)

        self.df = df

        exclude += df.select_dtypes(['datetime64[ns]', object]).columns.tolist()
        print(f"Exclude Fetures: {exclude}")

        if exclude:
            self.feats = df.columns.difference(exclude).tolist()
        else:
            self.feats = df.columns.tolist()

    def run(self):
        df = self.df.copy()

        with timer('干掉高缺失'):
            self.to_drop_dict['filter_missing'] = self.filter_missing()

        with timer('干掉低方差'):
            self.to_drop_dict['filter_variance'] = self.filter_variance()

        with timer('干掉高相关'):
            pass

        return df

    def filter_missing(self, feats=None, threshold=0.95):
        """
        :param feat_cols:
        :param threshold:
        :param as_na: 比如把-99当成缺失值
        :return:
        """
        if feats is None:
            feats = self.feats

        to_drop = self.df[feats].isna().mean()[lambda x: x > threshold].index.tolist()
        print('%d features with greater than %0.2f missing values.' % (len(to_drop), threshold))
        return to_drop

    def _filter_variance(self, feat, df):
        var = df[feat][lambda x: x.between(x.quantile(0.005), x.quantile(0.995))].var()
        return '' if var else feat

    def filter_variance(self, feats=None, max_worker=4):
        if feats is None:
            feats = self.feats

        _filter_variance = partial(self._filter_variance, df=self.df)
        with ProcessPoolExecutor(min(max_worker, len(feats))) as pool:
            to_drop = pool.map(_filter_variance, tqdm(feats, 'Filter Variance ...'))
            to_drop = [feat for feat in to_drop if feat]
        print('%d features with 0 variance in 0.5 ~ 99.5 quantile.' % len(to_drop))
        return to_drop


class FS(object):

    def __init__(self, estimator, verbose=0,
                 importance_type='split', importance_normlize=True,
                 percent=95, alpha=0.05, two_step=True,
                 ):
        self.estimator = estimator
        self.estimator_name = str(estimator).lower()
        self.verbose = verbose

        self.importance_type = importance_type  # split/shap/permutaion
        self.importance_type_normlize = importance_normlize

        self.alpha = alpha
        self.two_step = two_step
        self.percent = percent * 100 if 0 <= percent <= 1 else percent

    def fit(self, X, y, n_trials=10, sample_weight=None):

        # setup variables for Boruta
        n_sample, n_feat = X.shape

        # holds the decision about each feature:
        # 0  - default state = tentative in original code
        # 1  - accepted in original code
        # -1 - rejected in original code
        dec_reg = np.zeros(n_feat)

        # counts how many times a given feature was more important than the best of the shadow features
        hit_reg = np.zeros(n_feat)

        # these record the history of the iterations
        imp_history = np.zeros((n_trials, n_feat))  # 记录真实特征历史重要性, 输出排序
        imp_history[:] = np.nan

        # 影子特征历史重要性最大值
        sha_max_history = []

        # main feature selection loop
        for trial in tqdm(range(n_trials), desc='🔥'):
            if not np.any(dec_reg == 0):  # 存在未拒绝的就继续迭代
                break
            # make sure we start with a new tree in each iteration
            seed = np.random.randint(0, 10 ** 6)

            # 通过模型获取shap值
            imp_real, imp_sha = self._add_shadows_get_imps(X, y, dec_reg, sample_weight, seed=seed + 2)
            imp_history[trial] = imp_real  # record importance history

            # get the threshold of shadow importances we will use for rejection
            imp_sha_max = np.percentile(imp_sha, self.percent)
            sha_max_history.append(imp_sha_max)

            # register which feature is more imp than the max of shadows
            hits = np.where(imp_real > imp_sha_max)[0]
            hit_reg[hits] += 1

            # based on hit_reg we check if a feature is doing better than expected by chance
            dec_reg = self._do_tests(dec_reg, hit_reg, trial + 1)

        # we automatically apply R package's rough fix for tentative ones
        confirmed = np.where(dec_reg == 1)[0]
        tentative = np.where(dec_reg == 0)[0]

        # which tentative to keep
        tentative_median = np.median(imp_history[:, tentative], 0)
        tentative_confirmed = np.where(tentative_median > np.median(sha_max_history))[0]
        tentative = tentative[tentative_confirmed]

        # basic result variables
        self.n_features_ = confirmed.shape[0]
        self.support_ = np.zeros(n_feat, dtype=np.bool)
        self.support_[confirmed] = 1
        self.support_weak_ = np.zeros(n_feat, dtype=np.bool)
        self.support_weak_[tentative] = 1

        # ranking, confirmed variables are rank 1
        self.ranking_ = np.ones(n_feat, dtype=np.int)
        # tentative variables are rank 2
        self.ranking_[tentative] = 2
        # selected = confirmed and tentative
        selected = np.hstack((confirmed, tentative))
        # all rejected features are sorted by importance history
        not_selected = np.setdiff1d(np.arange(n_feat), selected)
        # large importance values should rank higher = lower ranks -> *(-1)
        imp_history_rejected = imp_history[:, not_selected] * -1

        # update rank for not_selected features
        if not_selected.shape[0] > 0:
            # calculate ranks in each iteration, then median of ranks across feats
            iter_ranks = self._nanrankdata(imp_history_rejected, axis=1)
            rank_medians = np.nanmedian(iter_ranks, axis=0)
            ranks = self._nanrankdata(rank_medians, axis=0)

            # set smallest rank to 3 if there are tentative feats
            if tentative.shape[0] > 0:
                ranks = ranks - np.min(ranks) + 3
            else:
                # and 2 otherwise
                ranks = ranks - np.min(ranks) + 2
            self.ranking_[not_selected] = ranks
        else:
            # all are selected, thus we set feature supports to True
            self.support_ = np.ones(n_feat, dtype=np.bool)

        self.importance_type_history_ = imp_history
        self.shadows_max_importance_history = sha_max_history

    def _estimator_fit(self, X_train, X_test, y_train, y_test, train_w=None, test_w=None, seed=0):
        # todo: oof estimator
        # check estimator_name
        self.estimator.set_params(random_state=seed)  # 要不要都行，因为数据随机了
        if 'lgb' in self.estimator_name:
            self.estimator.fit(X_train, y_train,
                               sample_weight=train_w,
                               eval_sample_weight=test_w,
                               eval_set=[(X_train, y_train), (X_test, y_test)],
                               eval_metric=None,
                               eval_names=('Train', 'Valid'),
                               verbose=self.verbose,
                               early_stopping_rounds=100)
        else:
            raise ValueError('默认支持lgb，其他模型待支持')

    def _add_shadows_get_imps(self, X, y, dec_reg, sample_weight=None, seed=0):
        # find features that are tentative still
        x_cur_ind = np.where(dec_reg >= 0)[0]
        x_cur = X[:, x_cur_ind]
        x_cur_w = x_cur.shape[1]

        # deep copy the matrix for the shadow matrix
        x_sha = x_cur.copy()

        # make sure there's at least 5 columns in the shadow matrix for
        while (x_sha.shape[1] < 5):
            x_sha = np.c_[x_sha, x_sha]

        # 打乱每列的值
        x_sha = np.apply_along_axis(check_random_state(seed).permutation, 0, x_sha)  # 打乱每列的值

        # get importance of the merged matrix
        imp = self._feature_importances(np.c_[x_cur, x_sha], y, sample_weight, seed + 1)  # (x_cur, x_sha)

        # separate importances of real and shadow features
        imp_sha = imp[x_cur_w:]

        imp_real = np.zeros(X.shape[1])
        imp_real[:] = np.nan
        imp_real[x_cur_ind] = imp[:x_cur_w]

        return imp_real, imp_sha

    def _feature_importances(self, X, y, sample_weight=None, seed=0):

        arrays = [X, y] if sample_weight is None else [X, y, sample_weight]
        args = train_test_split(*arrays, random_state=seed, stratify=y if check_classification(y) else None)
        self._estimator_fit(*args, seed=seed + 1)  # estimator fitted

        imp = get_imp(self.estimator, X, self.importance_type)

        return zscore(imp) if self.importance_type_normlize else imp

    def _nanrankdata(self, X, axis=1):
        """
        Replaces bottleneck's nanrankdata with scipy and numpy alternative.
        """
        ranks = sp.stats.mstats.rankdata(X, axis=axis)
        ranks[np.isnan(X)] = np.nan
        return ranks

    def _do_tests(self, dec_reg, hit_reg, _iter):
        active_features = np.where(dec_reg >= 0)[0]
        hits = hit_reg[active_features]
        # get uncorrected p values based on hit_reg
        to_accept_ps = sp.stats.binom.sf(hits - 1, _iter, .5).flatten()
        to_reject_ps = sp.stats.binom.cdf(hits, _iter, .5).flatten()

        if self.two_step:
            # two step multicor process
            # first we correct for testing several features in each round using FDR
            to_accept = self._fdrcorrection(to_accept_ps, alpha=self.alpha)[0]
            to_reject = self._fdrcorrection(to_reject_ps, alpha=self.alpha)[0]

            # second we correct for testing the same feature over and over again
            # using bonferroni
            to_accept2 = to_accept_ps <= self.alpha / float(_iter)
            to_reject2 = to_reject_ps <= self.alpha / float(_iter)

            # combine the two multi corrections, and get indexes
            to_accept *= to_accept2
            to_reject *= to_reject2
        else:
            # as in th original Boruta, we simply do bonferroni correction
            # with the total n_feat in each iteration
            to_accept = to_accept_ps <= self.alpha / float(len(dec_reg))
            to_reject = to_reject_ps <= self.alpha / float(len(dec_reg))

        # find features which are 0 and have been rejected or accepted
        to_accept = np.where((dec_reg[active_features] == 0) * to_accept)[0]
        to_reject = np.where((dec_reg[active_features] == 0) * to_reject)[0]

        # updating dec_reg
        dec_reg[active_features[to_accept]] = 1
        dec_reg[active_features[to_reject]] = -1
        return dec_reg

    def _fdrcorrection(self, pvals, alpha=0.05):
        """
        Benjamini/Hochberg p-value correction for false discovery rate, from
        statsmodels package. Included here for decoupling dependency on statsmodels.

        Parameters
        ----------
        pvals : array_like
            set of p-values of the individual tests.
        alpha : float
            error rate

        Returns
        -------
        rejected : array, bool
            True if a hypothesis is rejected, False if not
        pvalue-corrected : array
            pvalues adjusted for multiple hypothesis testing to limit FDR
        """
        pvals = np.asarray(pvals)
        pvals_sortind = np.argsort(pvals)
        pvals_sorted = np.take(pvals, pvals_sortind)
        nobs = len(pvals_sorted)
        ecdffactor = np.arange(1, nobs + 1) / float(nobs)

        reject = pvals_sorted <= ecdffactor * alpha
        if reject.any():
            rejectmax = max(np.nonzero(reject)[0])
            reject[:rejectmax] = True

        pvals_corrected_raw = pvals_sorted / ecdffactor
        pvals_corrected = np.minimum.accumulate(pvals_corrected_raw[::-1])[::-1]
        pvals_corrected[pvals_corrected > 1] = 1
        # reorder p-values and rejection mask to original order of pvals
        pvals_corrected_ = np.empty_like(pvals_corrected)
        pvals_corrected_[pvals_sortind] = pvals_corrected
        reject_ = np.empty_like(reject)
        reject_[pvals_sortind] = reject
        return reject_, pvals_corrected_



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : model_optimizers
# @Time         : 2021/9/14 下午12:39
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  :

import optuna
from lightgbm import LGBMClassifier


def LGBMClassifierTrial(params, trial: optuna.trial.Trial):
    opt_params = dict(
        num_leaves=trial.suggest_int("num_leaves", 2, 2 ** 8),
        learning_rate=trial.suggest_discrete_uniform('learning_rate', 0.001, 1, 0.001),
        n_estimators=trial.suggest_int("n_estimators", 2, 2 ** 10, log=True),

        min_child_samples=trial.suggest_int('min_child_samples', 2, 2 ** 8),
        min_child_weight=trial.suggest_loguniform('min_child_weight', 1e-8, 1),
        min_split_gain=trial.suggest_loguniform('min_split_gain', 1e-8, 1),

        subsample=trial.suggest_uniform('subsample', 0.4, 1),
        subsample_freq=trial.suggest_int("subsample_freq", 0, 2 ** 4),
        colsample_bytree=trial.suggest_uniform('colsample_bytree', 0.4, 1),
        reg_alpha=trial.suggest_loguniform('reg_alpha', 1e-8, 10),
        reg_lambda=trial.suggest_loguniform('reg_lambda', 1e-8, 10),
    )
    clf = LGBMClassifier(
        boosting_type='gbdt',
        num_leaves=31,
        max_depth=-1,
        learning_rate=0.1,
        n_estimators=100,
        subsample_for_bin=200000,
        objective=None,
        class_weight=None,
        min_split_gain=0.,
        min_child_weight=1e-3,
        min_child_samples=20,
        subsample=1.,
        subsample_freq=0,
        colsample_bytree=1.,
        reg_alpha=0.,
        reg_lambda=0.,
        random_state=None,
        n_jobs=-1,
        silent=True,
        importance_type='split'
    )
    clf.set_params(**{**opt_params, **params})
    return clf



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : model_utils
# @Time         : 2021/9/30 下午7:57
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *


def get_imp(estimator, X=None, importance_type=None):
    """
    todo: 采样加速，可多次采样取平均，近似全量数据的shap值
    """
    if importance_type is not None:
        importance_type = importance_type.lower()
    if importance_type in (None, 'tree', 'gini', 'split', 'gain'):
        imp = estimator.feature_importances_

    elif importance_type in 'shap':
        import shap
        explainer = shap.Explainer(estimator)
        shap_values = explainer(X).values
        imp = np.abs(shap_values).mean(0)

        if shap_values.ndim == 3:
            imp = imp.sum(1)

        # if isinstance(coefs, list):
        #     coefs = list(map(lambda x: np.abs(x).mean(0), coefs))
        #     coefs = np.sum(coefs, axis=0)
        # else:
        #     coefs = np.abs(coefs).mean(0)

    elif importance_type in 'permutaion':
        imp = ...  # todo

    else:
        raise ValueError('No Importances was specified select one of (shap, tree/gini/split, permutaion)')

    return imp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : gensim_w2v
# @Time         : 2021/9/13 上午9:34
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  : 增量学习 https://mp.weixin.qq.com/s/VIHwIjliwSJecmG6RVYTiA


from meutils.pipe import *
from collections import Iterable
from gensim.models import word2vec


class Word2Vec(object):
    """
    架构：skip-gram（慢、对罕见字有利）vs CBOW（快）
    训练算法：分层softmax（对罕见字有利）vs 负采样（对常见词和低纬向量有利）
    负例采样准确率提高，速度会慢，不使用negative sampling的word2vec本身非常快，但是准确性并不高
    欠采样频繁词：可以提高结果的准确性和速度（适用范围1e-3到1e-5）
    文本（window）大小：skip-gram通常在10附近，CBOW通常在5附近
    size: n = sqrt(词汇量)/2

    - [skip-gram][1]: 1个学生(中心词) `汇报 =>` K个老师(周围词)
    - cbow: K个学生(周围词) `汇报 =>` 1个老师(中心词)

    """

    def fit(self, corpus, vector_size=300, window=5, epochs=10, min_count=1, sg=0, hs=0, negative=5):
        """
        gensim.models.word2vec.Word2Vec?
        w2v.wv.key_to_index
        w2v.build_vocab(self.corpus, update=True) # 增量训练
        w2v.init_sims(replace=True) # 对model进行锁定，并且据说是预载了相似度矩阵能够提高后面的查询速度，但是你的model从此以后就read only了

        @param corpus: tqdm(sentences, desc="SentencePreprocessing")
        @param vector_size:
        @param window:
        @param min_count:
        @param sg:
        @param hs:
        @param negative:
        @param epochs:
        @return:
        """
        if isinstance(corpus, Iterable):
            sentences = corpus
            corpus_file = None

        else:  # Path(corpus).is_file()
            corpus_file = corpus
            sentences = None

        self.w2v = word2vec.Word2Vec(
            sentences,
            corpus_file,
            vector_size=vector_size,
            window=window,
            epochs=epochs,
            min_count=min_count,
            negative=negative,
            sg=sg, hs=hs,
            workers=32
        )

        logger.info(f"Word2Vec: ({len(self.w2v.wv.key_to_index)}, {vector_size})")


if __name__ == '__main__':
    docs = [['Well', 'done!'],
            ['Good', 'work'],
            ['Great', 'effort'],
            ['nice', 'work'],
            ['Excellent!'],
            ['Weak'],
            ['Poor', 'effort!'],
            ['not', 'good'],
            ['poor', 'work'],
            ['Could', 'have', 'done', 'better.']]
    w2v = Word2Vec()
    w2v.fit(docs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : encoders
# @Time         : 2021/9/23 上午11:11
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  :


from collections import OrderedDict

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class CountEncoder(BaseEstimator, TransformerMixin):

    def __init__(self, dropna=False, normalize=False):
        """

        :param dropna: 缺失值是否计数，默认计数
        :param normalize: 频数还是频率，默认频数
        """
        self.dropna = dropna
        self.normalize = normalize
        self.mapper = None

    def fit(self, y):
        self.mapper = (pd.Series(y).value_counts(self.normalize, dropna=self.dropna)
                       .to_dict(OrderedDict))
        return self

    def transform(self, y):
        """不在训练集的补0，不经常出现补0"""
        return pd.Series(y).map(self.mapper).fillna(0)


class CountRankEncoder(BaseEstimator, TransformerMixin):

    def __init__(self, topn=None):
        """

        :param topn: 仅保留topn个类别
        """
        self.topn = topn
        self.mapper = None

    def fit(self, y):
        ce = pd.Series(y).value_counts(True, dropna=False)  # 计数编码
        if self.topn:
            ce = ce[:self.topn]
            print(f"Coverage: {ce.sum() * 100:.2f}%")

        self.mapper = ce.rank(method='first').to_dict(OrderedDict)  # rank 合理？
        return self

    def transform(self, y):
        """不在训练集的补0，不经常出现补0"""
        return pd.Series(y).map(self.mapper).fillna(0)

class RankEncoder(BaseEstimator, TransformerMixin):

    def __init__(self, method='average', na_option='keep'):
        """
        method : {'average', 'min', 'max', 'first', 'dense'}, default 'average'
            How to rank the group of records that have the same value
            (i.e. ties):

            * average: average rank of the group
            * min: lowest rank in the group
            * max: highest rank in the group
            * first: ranks assigned in order they appear in the array
            * dense: like 'min', but rank always increases by 1 between groups
        numeric_only : bool, optional
            For DataFrame objects, rank only numeric columns if set to True.
        na_option : {'keep', 'top', 'bottom'}, default 'keep'
            How to rank NaN values:

            * keep: assign NaN rank to NaN values
            * top: assign smallest rank to NaN values if ascending
            * bottom: assign highest rank to NaN values if ascending
        ascending : bool, default True
            Whether or not the elements should be ranked in ascending order.
        """
        self.method = method
        self.na_option = na_option

    def transform(self, y):
        """不在训练集的补0，不经常出现补0"""
        return pd.Series(y).rank(method=self.method, na_option=self.na_option)  # .fillna(0)

if __name__ == '__main__':
    import numpy as np

    s = ['a', 'a', 'b', 'b', 'c'] + [np.nan] * 6
    re = CountRankEncoder()

    print(re.fit_transform(s))
    print(re.mapper)


    s = ['a', 'a', 'a', 'b', 'b'] + [np.nan] * 10
    ce = CountEncoder()
    print(ce.fit_transform(s))
    print(ce.mapper)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : encoders
# @Time         : 2020/10/2 2:48 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from collections import OrderedDict

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

from sklearn.preprocessing import LabelEncoder, OrdinalEncoder


class CountRankEncoder(object):

    def __init__(self, dropna=False):
        self.dropna = dropna

    def fit(self, y):
        self.mapper = (
            pd.Series(y).value_counts(dropna=self.dropna)
                .rank(method='first')
                .to_dict()
        )
        return self

    def transform(self, y):
        return pd.Series(y).map(self.mapper).fillna(0)


class CountEncoder(BaseEstimator, TransformerMixin):

    def __init__(self, dropna=False, normalize=False):
        """

        :param dropna: 缺失值是否计数，默认计数
        :param normalize: 频数还是频率，默认频数
        """
        self.dropna = dropna
        self.normalize = normalize
        self.mapper = None

    def fit(self, y):
        self.mapper = (
            pd.Series(y).value_counts(self.normalize, dropna=self.dropna)
                .to_dict(OrderedDict)
        )
        return self

    def transform(self, y):
        """不在训练集的补0，不经常出现补0"""
        return pd.Series(y).map(self.mapper).fillna(0)


class RankEncoder(BaseEstimator, TransformerMixin):

    def __init__(self, na_option='keep', ascending=True):
        """
        na_option : {'keep', 'top', 'bottom'}, default 'keep'
            How to rank NaN values:

            * keep: assign NaN rank to NaN values
            * top: assign smallest rank to NaN values if ascending
            * bottom: assign highest rank to NaN values if ascending
        ascending : bool, default True
            Whether or not the elements should be ranked in ascending order.
        """
        self.na_option = na_option
        self.ascending = ascending

    def transform(self, y):
        return pd.Series(y).rank(method='first', na_option=self.na_option, ascending=self.ascending)


if __name__ == '__main__':
    import numpy as np

    s = ['a', 'a', 'a', 'b', 'b'] + [np.nan] * 10
    ce = CountEncoder()
    print(ce.fit_transform(s))
    print(ce.mapper)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : oof
# @Time         : 2021/9/14 下午3:42
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  : todo: 增加nn模型


from sklearn.model_selection import ShuffleSplit, StratifiedKFold, train_test_split

# ME
from meutils.pipe import *
from aizoo.utils.model_utils import get_imp


class AdversarialValidation(object):
    """Adversarial Validation
    通过对抗验证确定最优拆分折数
    """

    def __init__(self, params=None, importance_type='split', fit_params=None):
        self.params = params if params is not None else {}
        self.params['metric'] = 'auc'

        self._importance_type = importance_type
        self.feature_importances_ = None

        self.fit_params = fit_params if fit_params is not None else {}

    def fit(self, X, y, cv=5, split_seed=777):
        self.feature_importances_ = np.zeros((cv, X.shape[1]))

        _ = enumerate(StratifiedKFold(cv, shuffle=True, random_state=split_seed).split(X, y))

        metrics = []
        for n_fold, (train_index, valid_index) in tqdm(_, desc='Train🔥'):
            print(f"\033[94mFold {n_fold + 1} started at {time.ctime()}\033[0m")
            # X_train, y_train = X[train_index], y[train_index]
            # X_valid, y_valid = X[valid_index], y[valid_index]
            X_train, X_valid = X[train_index], X[valid_index]

            y_ = np.zeros(len(X_train) + len(X_valid))
            y_[:len(X_train)] = 0
            y_[len(X_train):] = 1

            X_train, X_valid, y_train, y_valid = train_test_split(np.r_[X_train, X_valid], y_, stratify=y_)
            eval_set = [(X_train, y_train), (X_valid, y_valid)]

            import lightgbm as lgb

            estimator = lgb.LGBMClassifier()
            estimator.set_params(**self.params)

            fit_params = dict(
                eval_set=eval_set,
                eval_metric=None,
                eval_names=('Train🔥', 'Valid'),
                verbose=100,
                early_stopping_rounds=100,
            )

            estimator.fit(
                X_train, y_train,
                **{**fit_params, **self.fit_params}  # fit_params
            )
            metrics.append(estimator.best_score_['Valid']['auc'])

            # 记录特征重要性
            self.feature_importances_[n_fold] = get_imp(estimator, X_train, self._importance_type)

        _ = np.array(metrics)
        print(f"\n\033[94mValid: {_.mean():.6f} +/- {_.std():.6f} \033[0m\n")

        self.metrics = _  # auc、std 越小说明拆分的数据分布越接近

        return _.mean(), _.std()


class OOF(object):

    def __init__(self, params=None, fit_params=None, task='Classifier', importance_type='split'):
        """

        @param params:
        @param fit_params:
        @param weight_func:
        @param task: Classifier or Regressor
        """
        self.task = task.title()
        self.params = params if params is not None else {}
        self.fit_params = fit_params if fit_params is not None else {}

        self._estimators = []  # 每一折的模型
        self._importance_type = importance_type
        self.feature_importances_ = None

    @abstractmethod
    def fit_predict(self, X_train, y_train, w_train, X_valid, y_valid, w_valid, X_test, **kwargs):
        raise NotImplementedError

    @staticmethod
    def _predict(estimator):
        return estimator.predict_proba if hasattr(estimator, 'predict_proba') else estimator.predict

    def predict(self, X):

        func = lambda e: self._predict(e)(X)
        return np.array(list(map(func, self._estimators))).mean(0)

    def fit(self, X, y, sample_weight=None, X_test=None, feval=None, cv=5, split_seed=777, oof_file=None):

        if sample_weight is None:
            sample_weight = np.ones(len(X))

        X_test = X_test if X_test is not None else X[:66]

        self.feature_importances_ = np.zeros((cv, X.shape[1]))

        if self.task == 'Regressor':
            self.oof_train_proba = np.zeros(len(X))
            self.oof_test_proba = np.zeros(len(X_test))
            _ = enumerate(ShuffleSplit(cv, random_state=split_seed).split(X, y))  # todo: 兼容时间序列

        elif self.task == 'Classifier':
            num_classes = len(set(y))
            assert num_classes < 128, "是否是分类问题"
            self.oof_train_proba = np.zeros([len(X), num_classes])
            self.oof_test_proba = np.zeros([len(X_test), num_classes])
            _ = enumerate(StratifiedKFold(cv, shuffle=True, random_state=split_seed).split(X, y))
        else:
            raise ValueError("TaskTypeError⚠️")

        valid_metrics = []
        for n_fold, (train_index, valid_index) in tqdm(_, desc='Train 🐢'):
            print(f"\033[94mFold {n_fold + 1} started at {time.ctime()}\033[0m")
            X_train, y_train, w_train = X[train_index], y[train_index], sample_weight[train_index]
            X_valid, y_valid, w_valid = X[valid_index], y[valid_index], sample_weight[valid_index]

            ##############################################################
            valid_predict, test_predict = self.fit_predict(X_train, y_train, w_train, X_valid, y_valid, w_valid, X_test)
            ##############################################################

            self.oof_train_proba[valid_index] = valid_predict
            self.oof_test_proba += test_predict / cv

            # 记录特征重要性
            self.feature_importances_[n_fold] = get_imp(self._estimators[-1], X_train, self._importance_type)

            if feval is not None:
                if self.oof_test_proba.shape[-1] == 2:  # todo: 目前不支持多分类
                    valid_metrics.append(feval(y_valid, valid_predict[:, 1]))  # 二分类
                elif self.oof_test_proba.ndim == 1:
                    valid_metrics.append(feval(y_valid, valid_predict))  # 回归

        if self.oof_test_proba.shape[1] == 2:
            self.oof_train_proba = self.oof_train_proba[:, 1]
            self.oof_test_proba = self.oof_test_proba[:, 1]

        self.oof_train_test = np.r_[self.oof_train_proba, self.oof_test_proba]  # 方便后续stacking

        if feval is not None:
            self.oof_score = feval(y, self.oof_train_proba)

            print("\n\033[94mScore Info:\033[0m")
            print(f"\033[94m     {cv:>2} CV: {self.oof_score:.6f}\033[0m")

            _ = np.array(valid_metrics)
            print(f"\033[94m     Valid: {_.mean():.6f} +/- {_.std():.6f} \033[0m\n")

            return self.oof_score

        if oof_file is not None:
            pd.DataFrame({'oof': self.oof_train_test}).to_csv(oof_file, index=False)

    @classmethod
    def opt_cv(cls, X, y, X_test=None, cv_list=range(3, 16), params=None, **run_kwargs):
        """todo: 折数优化，看方差？"""

        scores = []
        for cv in tqdm(cv_list, desc='opt cv 🐢'):  # range(3, 16):
            oof = cls(params)
            _ = oof.fit(X, y, X_test, **run_kwargs)
            scores.append((_, cv, oof))

        return sorted(scores)[::-1]

    def plot_feature_importances(self, feature_names=None, topk=20, figsize=None, pic_name=None):
        import seaborn as sns
        import matplotlib.pyplot as plt

        columns = ['Importances', 'Features']
        importances = self.feature_importances_.mean(0)

        if feature_names is None:
            feature_names = list(map(lambda x: f'F_{x}', range(len(importances))))

        _ = sorted(zip(importances, feature_names), reverse=True)
        self.df_feature_importances = pd.DataFrame(_, columns=columns)

        plt.figure(figsize=(14, topk // 5) if figsize is None else figsize)
        sns.barplot(x=columns[0], y=columns[1], data=self.df_feature_importances[:topk])
        plt.title(f'Features {self._importance_type.title()} Importances')
        plt.tight_layout()

        if pic_name is not None:
            plt.savefig(f'importances_{self.oof_score}.png')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : __init__.py
# @Time         : 2021/9/23 上午11:08
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  :


from itertools import combinations

# ME
from meutils.pipe import *


class FE(object):

    def __init__(self, cat_feats=None, num_feats=None, cat_funcs=None, num_funcs=None, agg_dict=None,
                 cat_feat_combination=None):
        """

        @param cat_feats: cat_feats、num_feats同不为None
        @param num_feats:
        @param cat_funcs: ['nunique', 'max', 'min'] 已经包含count计算
        @param num_funcs: ['min', 'mean', 'median', 'max', 'sum', 'std', 'var', 'sem', 'skew']
        @param agg_dict: 覆盖默认的funcs
            {
                'N0': ['mean']
            }

        @cat_feat_combination: 自定义特征组合，默认所有组合 [['C0'], ...]

        """
        self.cat_feats = cat_feats
        self.num_feats = num_feats

        self.cat_funcs = cat_funcs if cat_funcs is not None else ['nunique']
        self.num_funcs = num_funcs if num_funcs is not None else ['mean', 'std']

        self.agg_dict = agg_dict if agg_dict is not None else {}

        self.cat_feat_combination = cat_feat_combination

    def transform(self, df: pd.DataFrame, n_jobs=1, only_features=False):
        """

        @param df:
        @param n_jobs:
        @return:
        """

        if self.cat_feats is None:
            self.cat_feats = self.infer_cat_feats(df)
            logger.info(f"Infer cat feats: {self.cat_feats}")

            if self.num_feats is None:
                self.num_feats = df.columns.difference(self.cat_feats).tolist()
                logger.info(f"Infer num feats: {self.num_feats}")

        if self.cat_feat_combination is None:
            self.cat_feat_combination = self.combination_all(self.cat_feats)

            ##############################################################
            # self.calculate_feature_number()
            ##############################################################

        if only_features:
            return self.cat_feat_combination  # 提前计算一下特征值数量，过滤

        bar = tqdm(self.cat_feat_combination, desc='agg🐢')

        if n_jobs == 1:
            for keys in bar:
                df_ = self.group_calculate(df, keys)
                df = df.merge(df_, 'left')  # df = df.merge(df_, on=keys)

        else:  # 多进程加速
            def func(keys):
                return self.group_calculate(df, keys)

            dfs = bar | xJobs(func, n_jobs) | xtqdm(desc=f'joining🐢')

            for df_ in dfs:
                df = df.merge(df_, 'left')  # todo: 并行设计

            #############################无加速效果#################################
            # def func(dfs):
            #     dff = df.copy()
            #     for df_ in dfs:
            #         dff = dff.merge(df_)
            #     return dff
            #
            # def dfs_join(l, func=func, n_jobs=n_jobs):
            #     batch_size = len(l) // n_jobs
            #
            #     if batch_size == 1:
            #         return func(l)
            #
            #     l = l | xgroup(batch_size) | xtqdm(desc=f'joining🐢bs={batch_size}') | xJobs(func, n_jobs)
            #
            #     return dfs_join(l)
            #
            # df = dfs_join(dfs)
            ##############################################################

        return df

    def group_calculate(self, df, keys):
        agg_dict = self._get_agg_dict(keys)
        agg_dict[keys[0]] = ['count']  # count只统计一次

        df_ = df.groupby(keys, dropna=False).agg(agg_dict)
        df_.columns = [f"{':'.join(keys)}.{func.upper()}({feat})" for feat, func in df_.columns]
        df_.reset_index(inplace=True)
        return df_

    def _get_agg_dict(self, keys):
        if self.num_funcs:
            num_agg_dict = {feat: self.num_funcs for feat in self.num_feats}
        else:
            num_agg_dict = {}  # num_funcs=[] 不聚合num特征

        if self.cat_funcs:
            cat_agg_dict = {feat: self.cat_funcs for feat in self.cat_feats if feat not in keys}
        else:
            cat_agg_dict = {}  # cat_funcs=[] 不聚合cat特征

        return {**cat_agg_dict, **num_agg_dict, **self.agg_dict}

    @staticmethod
    def combination_all(s):
        comb_list = []
        for i in range(1, len(s) + 1):
            comb_list += map(list, combinations(s, i))
        return comb_list

    @staticmethod
    def infer_cat_feats(df, threshold=64):
        threshold = min(len(df) // 100 + 8, threshold)
        return df.nunique()[lambda x: x < threshold].index.tolist()

    def calculate_feature_number(self):
        m = len(self.cat_feats)
        n = len(self.num_feats)

        cat_feat_num = ((2 ** m - 1) * m - sum(map(len, self.cat_feat_combination))) * len(self.cat_funcs) + \
                       len(self.cat_feat_combination)  # count特征数

        num_feat_num = (2 ** n - 1) * n * len(self.num_funcs)

        logger.info(f"CAT+NUM: {cat_feat_num + num_feat_num} = {cat_feat_num} + {num_feat_num}")


if __name__ == '__main__':
    df1 = pd.DataFrame({f"C{i}": np.random.randint(0, i + 3, size=100) for i in range(5)})
    df2 = pd.DataFrame(np.random.random((100, 5)), columns=(f"N{i}" for i in range(5)))

    df = pd.concat([df1, df2], 1)

    # ft = FE()
    # ft = FE(num_funcs=[])
    ft = FE(cat_funcs=[], num_funcs=[])

    df_ = ft.transform(df, n_jobs=2)

    print(df_.shape)

    print(df_.columns)

    print(ft.feat_value_num_)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : fm_embedding
# @Time         : 2020/10/9 8:19 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import numpy as np
from lightfm import LightFM

from scipy.sparse import csr_matrix


class FMEmbedding(object):

    def __init__(self, dim=128, **kwargs):
        self.fm = LightFM(no_components=dim, random_state=666, **kwargs)

    def fit(self, df, epochs=10, num_threads=30, verbose=True):
        """

        :param df: ['uid', 'iid', 'rating']
        :return:
        """
        df.columns = ['uid', 'iid', 'rating']
        csr_data = csr_matrix((df['rating'], (df['uid'], df['iid'])))

        print(f"csr_data: {csr_data.shape}")

        self.fm.fit(csr_data, epochs=epochs, num_threads=num_threads, verbose=verbose)

        self.user_embeddings = np.ascontiguousarray(self.fm.user_embeddings)
        self.item_embeddings = np.ascontiguousarray(self.fm.item_embeddings)

# if __name__ == '__main__':
#     import faiss
#     from ann import ANNFaiss
#
#     ann = ANNFaiss()
#
#     ann.train(item_embeddings, index_factory='IVF4000, Flat', metric=faiss.METRIC_INNER_PRODUCT, noramlize=True)
#
#     ann.noramlize()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : oof
# @Time         : 2021/9/14 下午3:42
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  : todo: 增加nn模型

# ME
from aizoo.base import OOF


class LGBMOOF(OOF):

    def fit_predict(self, X_train, y_train, w_train, X_valid, y_valid, w_valid, X_test, **kwargs):
        import lightgbm as lgb

        estimator = lgb.__getattribute__(f'LGBM{self.task}')()  # 实例
        estimator.set_params(**self.params)

        fit_params = dict(
            eval_set=[(X_train, y_train), (X_valid, y_valid)],
            eval_metric=None,
            eval_names=('Train🔥', 'Valid'),
            verbose=100,
            early_stopping_rounds=100,
            sample_weight=w_train,
            eval_sample_weight=[w_train, w_valid] # 与eval_set一致
        )

        estimator.fit(
            X_train, y_train,
            **{**fit_params, **self.fit_params}  # fit_params
        )

        self._estimators.append(estimator)
        predict = self._predict(estimator)
        return map(predict, (X_valid, X_test))


class XGBOOF(OOF):
    def fit_predict(self, X_train, y_train, w_train, X_valid, y_valid, w_valid, X_test, **kwargs):
        import xgboost as xgb

        estimator = xgb.__getattribute__(f'XGB{self.task}')()  # 实例
        estimator.set_params(**self.params)

        fit_params = dict(
            eval_set=[(X_train, y_train), (X_valid, y_valid)],
            eval_metric=None,
            verbose=100,
            early_stopping_rounds=100,
            sample_weight=w_train,
            sample_weight_eval_set=[w_train, w_valid]  # 与eval_set一致
        )

        estimator.fit(
            X_train, y_train,
            **{**fit_params, **self.fit_params}
        )

        self._estimators.append(estimator)
        predict = self._predict(estimator)
        return map(predict, (X_valid, X_test))


class CatBoostOOF(OOF):

    def fit_predict(self, X_train, y_train, w_train, X_valid, y_valid, w_valid, X_test, **kwargs):
        import catboost as cab
        estimator = cab.__getattribute__(f'CatBoost{self.task}')()  # TODO: embedding_features
        estimator.set_params(**self.params)

        fit_params = dict(
            eval_set=(X_valid, y_valid),  # CatBoostError: Multiple eval sets are not supported on GPU
            verbose=100,
            early_stopping_rounds=100,
            sample_weight=w_train,

            use_best_model=True,
            plot=True,
        )

        estimator.fit(
            X_train, y_train,
            **{**fit_params, **self.fit_params}
        )
        # evals_result = estimator.evals_result()
        self._estimators.append(estimator)
        predict = self._predict(estimator)
        return map(predict, (X_valid, X_test))


class TabNetOOF(OOF):

    def fit_predict(self, X_train, y_train, w_train, X_valid, y_valid, w_valid, X_test, **kwargs):
        if self.task == 'Regressor':  # tabnet 回归输入的不同
            y_train = y_train.reshape(-1, 1)
            y_valid = y_valid.reshape(-1, 1)

        from pytorch_tabnet import tab_model

        estimator = tab_model.__getattribute__(f'TabNet{self.task}')()  # TODO: embedding_features
        estimator.set_params(**self.params)

        fit_params = dict(
            eval_set=[(X_train, y_train), (X_valid, y_valid)],
            eval_name=('Train🔥', 'Valid'),
            eval_metric=None,
            max_epochs=100,
            patience=5
        )

        estimator.fit(
            X_train, y_train,
            **{**fit_params, **self.fit_params}
        )
        self._estimators.append(estimator)
        predict = self._predict(estimator)

        if self.task == 'Regressor':
            return predict(X_valid).reshape(-1), predict(X_test).reshape(-1)
        else:
            return map(predict, (X_valid, X_test))


if __name__ == '__main__':
    import shap
    from sklearn.metrics import r2_score, roc_auc_score
    from sklearn.datasets import make_regression, make_classification

    for Model in [LGBMOOF]:
        X, y = make_classification(n_samples=1000)
        oof = Model(importance_type='split')
        oof.fit(X, y, feval=roc_auc_score, cv=5)
        print(oof.predict(X))

        break

    # for Model in [LGBMOOF, TabNetRegressor, CatBoostRegressor, , XGBRegressor]:
    #     X, y = make_regression(n_samples=1000)
    #     oof = Model(weight_func=lambda w: 1 / (w + 1))
    #     oof.run(X, y, feval=r2_score, cv=5)
    #     break



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : optimizers
# @Time         : 2021/9/16 上午11:00
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  :


from meutils.pipe import *
from sklearn.metrics import *

from aizoo.tuner.base import optuna, Tuner
from aizoo.tab.models import LGBMOOF


class F1Optimizer(Tuner):

    def __init__(self, search_space, y, y_pred, **kwargs):
        super().__init__(search_space, **kwargs)
        self.y = y
        self.y_pred = y_pred

    def objective(self, trial: optuna.trial.Trial):
        params = self.trial_choice(trial)

        y_pred_ = np.where(np.array(self.y_pred) > params['threshold'], 1, 0)
        return f1_score(self.y, y_pred_)


class LGBOptimizer(Tuner):

    def __init__(self, search_space, X, y, feval=roc_auc_score, **kwargs):
        super().__init__(search_space, **kwargs)
        self.X = X
        self.y = y
        self.feval = feval

    def objective(self, trial: optuna.trial.Trial, task='Classifier'):
        params = self.trial_choice(trial)
        _ = LGBMOOF(params, task=task).fit(self.X, self.y, feval=self.feval)
        return _


if __name__ == '__main__':
    y = [1, 1, 0, 0]
    y_pred = [0.1, 0.2, 0.3, 0.4]

    # opt = F1Optimizer({'threshold': 0.1}, y, y_pred)
    opt = F1Optimizer("./search_space/f1.yaml", y, y_pred)

    opt.optimize(
        100,
        direction='minimize',
        study_name='test',
        storage="sqlite:////Users/yuanjie/Desktop/Projects/Python/aizoo/aizoo/tuner/test.db",
        load_if_exists=True  # cli --skip-if-exists
    )
    # optuna-dashboard sqlite:////Users/yuanjie/Desktop/Projects/Python/aizoo/aizoo/tuner/test.db

    print(opt.trials_dataframe)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : aizoo.
# @File         : check_utils
# @Time         : 2021/9/27 下午9:18
# @Author       : yuanjie
# @WeChat       : 313303303
# @Software     : PyCharm
# @Description  :


from meutils.pipe import *


def check_classification(y, threshold=128):
    """

    @param y:
    @param threshold: 目标值多于 threshold 个就认为是回归问题
    @return:
    """
    return len(set(y)) < threshold


def check_pandas_input(self, arg):
    try:
        return arg.values
    except AttributeError:
        raise ValueError(
            "input needs to be a numpy array or pandas data frame."
        )


